/*--------------------------------------------------------------*/
/* config.c -- .cfg file read/write for route                  	*/
/*--------------------------------------------------------------*/
/*                Personal work of Steve Beccue                 */
/*            Copyright (C) 2003 - All Rights Reserved          */
/*--------------------------------------------------------------*/
/* Modified by Tim Edwards, June 2011.  The route.cfg file is	*/
/* no longer the main configuration file but is supplementary	*/
/* to the LEF and DEF files.  Various configuration items that	*/
/* do not appear in the LEF and DEF formats, such as route	*/
/* costing, appear in this file, as well as a filename pointer	*/
/* to the LEF file with information on standard cell macros.	*/
/*--------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "qrouter.h"
#include "config.h"

int    CurrentPin = 0;
int    Firstcall = TRUE;
int    PinNumber = 0;

int     Num_layers   = MAX_LAYERS; // layers to use to route

double  PathWidth[MAX_LAYERS];     // width of the paths
int     GDSLayer[MAX_LAYERS];	   // GDS layer number 
int     GDSCommentLayer = 1;	   // for dummy wires, etc.
char    CIFLayer[MAX_LAYERS][50];  // CIF layer name
double  PitchX[MAX_LAYERS];         // Horizontal wire pitch of layer
double  PitchY[MAX_LAYERS];         // Vertical wire pitch of layer
int     NumChannelsX[MAX_LAYERS];  // number of wire channels in X on layer
int     NumChannelsY[MAX_LAYERS];  // number of wire channels in Y on layer
int     Vert[MAX_LAYERS];          // 1 if vertical, 0 if horizontal
int     Numpasses = 1;             // number of times to iterate in route_segs

double  Xlowerbound=0.0;   // Bounding Box of routes, in microns
double  Xupperbound=0.0;      
double  Ylowerbound=10.0;
double  Yupperbound=10.0;      

int     SegCost = 1;               // Route cost of a segment
int     ViaCost = 3;               //  "  of via between adjacent layers
int     JogCost = 5;               //  cost of 1 grid off-direction jog
int     XverCost = 0;              //  cost of a crossover
int     BlockCost = 100;           //  cost of a crossover when node has
				   //  only one tap point

char    Via[MAX_LAYERS][MAX_NAME_LEN];

/*--------------------------------------------------------------*/
/* read_config - read in the config file        		*/
/*								*/
/*         ARGS: the filename (normally route.cfg)		*/
/*      RETURNS: number of lines successfully read		*/
/* SIDE EFFECTS: loads Global configuration variables		*/
/*--------------------------------------------------------------*/

#define MAXLINE    256

int read_config(FILE *configfileptr)
{
    int count, lines, i, OK, h, v;
    int iarg, iarg2;
    char carg, gatename[MAX_NAME_LEN];
    double darg, darg2, darg3;
    char sarg[MAXLINE];
    FILE * cin;
    size_t len = 0;
    char   line[MAXLINE];
    STRING dnr;               // Do Not Route nets
    STRING cn;                // critical nets
    STRING strl;
    GATE   gateinfo;          // gate information, pin location, etc
    DSEG   grect, drect;

    if (Firstcall) {
	for (i = 0; i < MAX_LAYERS; i++) {
	    sprintf(Via[i], "via%d%d", i + 1, i + 2);
	}

	DontRoute = (STRING)NULL;
	CriticalNet = (STRING)NULL;
        GateInfo = (GATE)NULL;
	Nlgates = (GATE)NULL;

	for (i = 0; i < MAX_LAYERS; i++)
	   PitchX[i] = PitchY[i] = 0.0;

	Firstcall = 0;
    }

    cin = configfileptr;
    if (!cin) {
	return -1;
    }

    count = 0;
    lines = 0;

    while (!feof(cin)) {
	fgets(line, MAXLINE, cin);
	lines++;

	for (i = 0; line[i] && i < MAXLINE - 1; i++) {
	    line[i] = (char)tolower(line[i]);
	}
	
	if ((i = sscanf(line, "num_layers %d", &iarg)) == 1) {
	    OK = 1; Num_layers = iarg;
	}

	if ((i = sscanf(line, "layer_%d_name %s", &iarg2, sarg)) == 2) {
	    if (iarg2 > 0 && iarg2 < 10) {
	       OK = 1; strcpy(CIFLayer[iarg2 - 1], sarg);
	    }
	}

	if ((i = sscanf(line, "gds_layer_%d %d", &iarg2, &iarg)) == 2) {
	    if (iarg2 > 0 && iarg2 < 10) {
	        OK = 1; GDSLayer[iarg2 - 1] = iarg;
	    }
	}
	if ((i = sscanf(line, "gds_comment_layer %d", &iarg)) == 1) {
	    OK = 1; GDSCommentLayer = iarg;
	}

	if ((i = sscanf(line, "layer_1_width %lf", &darg)) == 1) {
	    OK = 1; PathWidth[0] = darg;
	}
	if ((i = sscanf(line, "layer_2_width %lf", &darg)) == 1) {
	    OK = 1; PathWidth[1] = darg;
	}
	if ((i = sscanf(line, "layer_3_width %lf", &darg)) == 1) {
	    OK = 1; PathWidth[2] = darg;
	}
	if ((i = sscanf(line, "layer_4_width %lf", &darg)) == 1) {
	    OK = 1; PathWidth[3] = darg;
	}
	if ((i = sscanf(line, "layer_5_width %lf", &darg)) == 1) {
	    OK = 1; PathWidth[4] = darg;
	}
	if ((i = sscanf(line, "layer_6_width %lf", &darg)) == 1) {
	    OK = 1; PathWidth[5] = darg;
	}
	if ((i = sscanf(line, "layer_7_width %lf", &darg)) == 1) {
	    OK = 1; PathWidth[6] = darg;
	}
	if ((i = sscanf(line, "layer_8_width %lf", &darg)) == 1) {
	    OK = 1; PathWidth[7] = darg;
	}
	if ((i = sscanf(line, "layer_9_width %lf", &darg)) == 1) {
	    OK = 1; PathWidth[8] = darg;
	}

	if ((i = sscanf(line, "x lower bound %lf", &darg)) == 1) {
	    OK = 1; Xlowerbound = darg;
	}
	if ((i = sscanf(line, "x upper bound %lf", &darg)) == 1) {
	    OK = 1; Xupperbound = darg;
	}
	if ((i = sscanf(line, "y lower bound %lf", &darg)) == 1) {
	    OK = 1; Ylowerbound = darg;
	}
	if ((i = sscanf(line, "y upper bound %lf", &darg)) == 1) {
	    OK = 1; Yupperbound = darg;
	}
	
	if ((i = sscanf(line, "layer %d wire pitch %lf\n", &iarg, &darg)) == 2) {
	    OK = 1; PitchX[iarg-1] = darg;
	}
	else if (i == 1) {
	   if ((i = sscanf(line, "layer %*d vertical %d\n", &iarg2)) == 1) {
	      OK = 1; Vert[iarg - 1] = iarg2;
	   }
	   else if ((i = sscanf(line, "layer %*d %c\n", &carg)) == 1) {
	      if (tolower(carg) == 'v') {
		 OK = 1; Vert[iarg - 1] = 1;
	      }
	      else if (tolower(carg) == 'h') {
		 OK = 1; Vert[iarg - 1] = 0;
	      }
	   }
	}

	if ((i = sscanf(line, "num passes %d\n", &iarg)) == 1) {
	    OK = 1;
	    Numpasses = iarg;
	}
	
	if ((i = sscanf(line, "route segment cost %d", &iarg)) == 1) {
	    OK = 1; SegCost = iarg;
	}
	
	if ((i = sscanf(line, "route via cost %d", &iarg)) == 1) {
	    OK = 1; ViaCost = iarg;
	}
	
	if ((i = sscanf(line, "route jog cost %d", &iarg)) == 1) {
	    OK = 1; JogCost = iarg;
	}
	
	if ((i = sscanf(line, "route crossover cost %d", &iarg)) == 1) {
	    OK = 1; XverCost = iarg;
	}

	if ((i = sscanf(line, "route block cost %d", &iarg)) == 1) {
	    OK = 1; BlockCost = iarg;
	}

	if ((i = sscanf(line, "do not route node %s\n", sarg)) == 1) {
	    OK = 1; 
	    dnr = (STRING)malloc(sizeof(struct string_));
	    dnr->name = strdup(sarg);
	    if (DontRoute != NULL) {
	       for (strl = DontRoute; strl->next; strl = strl->next);
	       strl->next = dnr;
	    }
	    else {
	       dnr->next = NULL;
	       DontRoute = dnr;
	    }
	}
	
	if ((i = sscanf(line, "route priority %s\n", sarg)) == 1) {
	    OK = 1; 
	    cn = (STRING)malloc(sizeof(struct string_));
	    cn->name = strdup(sarg);
	    if (CriticalNet != NULL) {
	       for (strl = CriticalNet; strl->next; strl = strl->next);
	       strl->next = cn;
	    }
	    else {
	       cn->next = NULL;
	       CriticalNet = cn;
	    }
	}
	
	if ((i = sscanf(line, "critical net %s\n", sarg)) == 1) {
	    OK = 1; 
	    cn = (STRING)malloc(sizeof(struct string_));
	    cn->name = strdup(sarg);
	    if (CriticalNet != NULL) {
	       for (strl = CriticalNet; strl->next; strl = strl->next);
	       strl->next = cn;
	    }
	    else {
	       cn->next = NULL;
	       CriticalNet = cn;
	    }
	}
	if ((i = sscanf( line, "lef %s\n", sarg)) == 1) {
	    // Argument is a filename of a LEF file from which we
	    // should get the information about gate pins & obstructions
	    OK = 1;
	    LefRead(sarg);
	}
	
	if ((i = sscanf(line, "gate %s %lf %lf\n", sarg, &darg, &darg2)) == 3) {
	    OK = 1; 
	    CurrentPin = 0;
	    gateinfo = (GATE)malloc(sizeof(struct gate_));
	    gateinfo->gatetype = strdup(sarg);
	    gateinfo->gatename = NULL;
	    gateinfo->width = darg;
	    gateinfo->height = darg2;
	    gateinfo->placedX = 0.0;	// implicit cell origin
	    gateinfo->placedY = 0.0;
	    gateinfo->next = GateInfo;	// prepend to linked gate list
	    GateInfo = gateinfo;
	}
	
        if ((i = sscanf(line, "endgate %s\n", sarg)) == 1) {
	    OK = 1; 
	    gateinfo->nodes = CurrentPin;

	    // This syntax does not include declaration of obstructions
	    gateinfo->obs = (DSEG)NULL;
	    CurrentPin = 0;
        }

	if ((i = sscanf(line, "pin %s %lf %lf\n", sarg, &darg, &darg2)) == 3) {
	    OK = 1; 
	    gateinfo->node[CurrentPin] = strdup(sarg);

	    // These style gates have only one tap per gate;  LEF file reader
	    // allows multiple taps per gate node.

	    drect = (DSEG)malloc(sizeof(struct dseg_));
	    gateinfo->taps[CurrentPin] = drect;
	    drect->x1 = drect->x2 = darg;
	    drect->y1 = drect->y2 = darg2;

	    // This syntax always defines pins on layer 0;  LEF file reader
	    // allows pins on all layers.

	    drect->layer = 0;
	    drect->next = (DSEG)NULL;
	    CurrentPin++;
	}

	if (OK == 0) {
	    if (!(line[0] == '\n' || line[0] == '#' || line[0] == 0)) {
		fprintf(stderr, "line not understood: %s\n", line);
	    }
	}
	OK = 0;
	line[0] = line[1] = 0;
    }

    // The following code ensures that the layer grids align.
    // For now, all PitchX[i] and PitchY[i] should be the same
    // for all layers.  Hopefully this restriction can be lifted
    // sometime, but it will necessarily be a royal pain.

    h = v = -1;
    for (i = 0; i < Num_layers; i++) {
       if (!Vert[i]) {
	  h = i;
	  PitchY[i] = PitchX[i];
	  PitchX[i] = 0.0;
       }
       else
	  v = i;
    }

    // In case all layers are listed as horizontal or all
    // as vertical, we should still handle it gracefully

    if (h == -1) h = v;
    else if (v == -1) v = h;

    for (i = 0; i < Num_layers; i++) {
       if (PitchX[i] != 0.0 && PitchX[i] != PitchX[v])
	  fprintf(stderr, "Multiple vertical route layers at different"
		" pitches.  I cannot handle this!  Using pitch %g\n",
		PitchX[v]);
       PitchX[i] = PitchX[v];
       if (PitchY[i] != 0.0 && PitchY[i] != PitchY[h])
	  fprintf(stderr, "Multiple horizontal route layers at different"
		" pitches.  I cannot handle this!  Using pitch %g\n",
		PitchY[h]);
       PitchY[i] = PitchY[h];
    }
    return count;

} /* read_config() */

